
 package Ali100;

################################################################
###                                                          ###
### SCaFoS                                                   ###
### Lecture des fichiers de sequences alignees               ###
###                                                          ###
################################################################
### version 1.00 du 2005/12/09                           BPH ###
################################################################


 use strict;
 use warnings;
 use Cwd;
 
 use Var100;
 use Lib100;
 use Gub100;

 use vars qw(@ISA @EXPORT);
 require Exporter;
 @ISA = qw(Exporter);
 @EXPORT = qw/ LoadFileAli SearchFileType ForbidenChar /;

 
################################################################


 my $L;
 my $bAff = '';       # texte pour affichage en mode graphique
 

################################################################
### Lecture des fichiers de sequences alignees #################

# NOM : ForbidenChar 
# DESCRIPTION : verifie que la chaine ne contient pas des caracteres interdits
# PARAMETRES : chaine a tester
# RETOUR : 0 -> OK      1 -> existence caracteres interdits
sub ForbidenChar {
    my ($Name) = @_;
    my $b;
    my $ret = 0;
    my $n=0;
 
    $b = $Name;
    $n++ while ( $b   =~ m/\\|\||\(|\)|\[|\]|\{|\}|\^|\$|\*|\+|\?|~|!|%|&|=|'|;|:|,|<|>|]/g);
    if ( $n > 0 ) { $ret = 1; }
    return($ret);
}

 
# NOM : LoadFileAli 
# DESCRIPTION : recupere tous les fichers du repertoire en entree
# PARAMETRES : so
# RETOUR : so
sub LoadFileAli {
    my @T1;
    $TypeSeq = '';
    if ( SearchFileType() eq '' ) { return(-1); };
    opendir REP, $FileRep;
    @T1 = grep m/.ali$/i, readdir REP;
    closedir REP;
    if ( ReadFileAli(@T1) == -1 ) { return(-1); };
    if ( $NbFileAli == 0 ) {
        $bAff = "None correct input file in the directory \'$FileRep\'";
        if ( $FlagIHM == 1 ) { return(ViewInfo('Warning' , $bAff, 'blue', $FenCur)); }
        else { die("\n $bAff\n\n"); }
        }
    return(0);
}


#### 4.03 v
# NOM : SearchFileType 
# DESCRIPTION : determine le type de fichier en entree
# PARAMETRES : so
# RETOUR : type de fichiers
sub SearchFileType {
    my $F;
    my $flagInter;
    my $nbseq;
    my $lgseq;
    my $flagCut;
    my $flagblock;
    my $nbBlocks;
    my $lgBlock;
    my @SeqFile;
    my @T1;
    my $SeqName = '';
    my $Seq = '';
    my $NbType = 0;
    
    #-- Type de fichiers ---------------------------------------
    $TypeAli = '';
    #-- MUST ---
    if ( !(opendir REP, $FileRep) )
        { AffError("Can\'t open the directory \'$FileRep\'"); }
    @T1 = grep m/ali$/i, readdir REP;
    closedir REP;
    if ( $#T1 >= 0 ) {
        $TypeAli = '.ali';
        $NbType += 1;
        }
        
    #-- FASTA ---
    opendir REP, $FileRep;
    @T1 = grep m/fasta$/i, readdir REP;
    closedir REP;
    if ( $#T1 >= 0 ) {
        $TypeAli = '.fasta';
        $NbType += 1;
        }
    
    #-- PHYLIP ---
    opendir REP, $FileRep;
    @T1 = grep m/phylip$/i, readdir REP;
    closedir REP;
    if ( $#T1 >= 0 ) { 
        $TypeAli = '.phylip'; 
        $NbType += 1;
        }
    
    #-- 1 et 1 seul format ---
    if ( $NbType == 0 ) {
        $bAff = "None correct input file in the directory \'$FileRep\'";
        if ( $FlagIHM == 1 ) { return(ViewInfo('Warning' , $bAff, 'blue', $FenCur)); }
        else { die "\n $bAff\n\n"; }
        }
    elsif ( $NbType > 1 ) {
        $bAff = "All input files must have the same format in the directory \'$FileRep\'";
        if ( $FlagIHM == 1 ) { return(ViewInfo('Warning' , $bAff, 'blue', $FenCur)); }
        else { die "\n $bAff\n\n"; }
        }

    #-- fichiers FASTA : decompose et copie en .ali -------------------------
    if ( $TypeAli eq '.fasta' ) {
        opendir REP, $FileRep;
        @T1 = grep m/fasta$/i, readdir REP;
        closedir REP;
        foreach my $F (sort @T1) {
            # lecture du .fasta
            if ( !open(FIC, $FileRep.$SepDir.$F) )
                { AffError("Can\'t open tfile \'$FileRep".'/'.$T1[0]."\'"); }
            if ( !open(ALI, ">$FileRep".$SepDir."$F.ali") )
                { AffError("Can\'t open file \'$FileRep\/$F.ali\' for writing"); }
            $Seq = "";
            $SeqName = "";
            while ( $L = <FIC> ) {
                $L = EndLine($L);
                if ( length($L) == 0 or $L =~ m/#/ ) { next; }
                if ( $L =~ m/^>(.*)/ ) {
                    # stocke sequence precedente
                    if ( $SeqName ne '' ) {
####                        $Seq =~ s/\s//g;
                        $Seq =~ s/\s+//g;
                        print ALI ">$SeqName\n$Seq\n";
                        }
                    $Seq = "";
                    # recupere le nom de l'espece suivante
                    $SeqName = $1;
                    }
                elsif ( length($SeqName) > 0 )
                    { $Seq .= $L; }
                }
            close FIC;
            if ( $SeqName ne '' and $Seq ne '' ) {
####                $Seq =~ s/\s//g;
                $Seq =~ s/\s+//g;
                print ALI ">$SeqName\n$Seq\n";
                }
            close ALI;
            }
        }

    #-- fichier PHYLIP : decompose et copie en .ali ------------
    elsif ( $TypeAli eq '.phylip' ) {
        my ($i, $j);
        my $bFile;
        my $F;
        my $nameseq;
        my $type = '';
        my %hSeq = ();
        
        opendir REP, $FileRep;
        @T1 = grep m/phylip$/i, readdir REP;
        closedir REP;
        foreach $F (sort @T1) {
            if ( !open(FIC, $FileRep.$SepDir.$F) )
                { AffError("Can\'t open file \'$FileRep".'/'.$T1[0]."\'"); }
            #-- traite la 1ere ligne pour recuperer nb et lgr sequences ---
            $L = <FIC>;
            if ( !( $L =~ m/\s*(\d*)\s*(\d*)\s*(\w*)/ ) )
                { AffError("None correct PHYLIP format for file \'$FileRep/$F\'"); }
            $nbseq = $1;
            $lgseq = $2;
            $type = $3;
            
            #-- Verifie le format du fichier ---
            # lit 1ere ligne de sequence
            $L = <FIC>;
            chomp($L);
            $flagInter = 0;
            $nameseq = substr($L, 0, 10);
            $L = substr($L, 10);
            if ( $L =~ m/(\D*)(\d*)\s*$nameseq\w*/ ) { $L = $1; }
####            $L =~ s/\s//g;
            $L =~ s/\s+//g;
            $lgBlock = length($L);
            # lit 2eme ligne pour voir si nouvelle sequence ou suite de la precedente
            $L = <FIC>;
            chomp($L);
            $flagCut = 0;
            $nbBlocks = 0;    # 4.10-b
            if ( substr($L, 0, 10) eq '          ' )
                # suite de la meme sequence
                { $flagCut = 1; }
            else {
                # nouvelle sequence
                if ( $lgBlock < $lgseq ) {
                    $flagInter = 1;
                    $nbBlocks = int($lgseq / $lgBlock);
                    }
                $nbBlocks += 1;
                }
            close FIC;

            #-- lit le fichier du debut et recupere les sequences ---
            open(FIC, $FileRep.$SepDir.$F);
            # passe 1ere ligne du fichier
            $L = <FIC>;
                
            # sequences sequentielles sur plusieurs lignes
            $bFile = '';
            $flagblock = 0;
            if ( $flagCut == 1 ) {
                while ( $L = <FIC> ) {
                    chomp($L);
                    if ( length($L) == 0 ) { next; }
                    # recupere nom sequence
                    if ( substr($L,0,10) ne '          ' ) {
                        $nameseq = substr($L,0,10);
                        $bFile .= "\n>$nameseq\n";
                        $L = substr($L, 10);
                        }
                    # recupere sequence
                    if ( $L =~ m/(\D*)(\d*)\s*$nameseq\w*/ ) { $L = $1; }
####                    $L =~ s/\s//g;
                    $L =~ s/\s+//g;
                    $bFile .= $L;
                    }
                }
                
            # sequences sur une seule ligne ou intercalees
            else {
                @SeqFile = ();
                for ($j=0; $j<$nbBlocks; $j++) {
                    for ( $i=0; $i<$nbseq; $i++ ) {
                        $L = <FIC>;
                        chomp($L);
                        # ligne vide : on passe a la suivante
                        if ( $L eq '')
                            { $i -= 1; }
                        # 1er bloc : on recupere le nom et la sequence
                        elsif ( $j == 0 and $L ne '') {
                            $nameseq = substr($L,0,10);
                            push @SeqFile, ">$nameseq\n";
                            $L = substr($L, 10);
                            if ( $L =~ m/(\D*)(\d*)\s*$nameseq\w*/ ) { $L = $1; }
####                            $L =~ s/\s//g;
                            $L =~ s/\s+//g;
                            $L = substr($L, 0, $lgBlock);
                            $SeqFile[$i] .= $L;
                            }
                        # on recupere la sequence
                        else {
                            if ( $L =~ m/(\D*)(\d*)\s*$nameseq\w*/ ) { $L = $1; }
####                            $L =~ s/\s//g;
                            $L =~ s/\s+//g;
                            $L = substr($L, 0, $lgBlock);
                            $SeqFile[$i] .= $L;
                            }
                        }
                    }
                for $L (@SeqFile) { $bFile .= "$L\n"; }
                }
            close FIC;
            #-- reecrit le fichier au format MUST ---
            if ( !open(FIC, ">$FileRep".$SepDir."$F.ali") )
                { AffError("Can\'t open file \'$FileRep\/$F.ali\' for writing"); }
            print FIC $bFile;
            close FIC;
            }
        }
        
    return($TypeAli);
}


### Verifie que tous les fichiers contiennent le meme type de sequences ### 
# NOM : TestSeq 
# DESCRIPTION : Verifie que tous les fichiers contiennent le meme type de sequences
# PARAMETRES : Type du fichier precedent, Sequence a tester
# RETOUR : Type du fichier (protein ou DNA)
sub TestSeq {
    my ($FlagType, $Seq) = @_;
    my $flag = 0;
    my $bAff;
    
    #1er appel : definit le type de sequence
    if ( $FlagType eq '' ) {
        if ( $Seq =~ m/[EFILPQ]/ )
            { $FlagType = 'protein'; }
        else
            { $FlagType = 'DNA'; }
        }
    #appels suivants: verifie que le type ne change pas
    else {
        if ( $Seq =~ m/[EFILPQ]/ )
            { if ( $FlagType ne 'protein' ) { $flag = -1; } }
        else
            { if ( $FlagType ne 'DNA' ) { $flag = -1; } }
        }
    if ( $flag == -1 )
        { AffError("All aligned files must contain the same type of sequences (proteins ou nucleic acids)"); }
    return($FlagType);
}


# NOM : ReadFileAli 
# DESCRIPTION : decompose les fichiers par espece
# PARAMETRES : Liste des fichiers
# RETOUR : so
sub ReadFileAli {
    my @T1 = @_;
    my $FicOut = "";    # liste des fichiers avec au moins 2 seq avec noms identiques
    my $fOut = 0;       # drapeau pour fin d'execution du programme
    my $Choose = "";    # choix de l'utilisateur pour modifier noms sequences
    my $nbBlancs;
    my $b = '';
    my $ListEsp = "<<<<<";
    my $buf;
    my $LSpe;
    my $SpeOut;
    my $Lg;
    my @T2;
    my @NewFic;
    my $n;
    my @tName;
    my $SeqName;
    my $E;
    my $L2;
    my $Seq;
    my $PercentLen;
    my $Result;
#    my $TypeSeq = '';    # protein ou DNA/RNA
    my $flag;
    my $FlagDiese;
    my $ListForbiden;
    my $FlagForbiden;
    my $b1;
    
    $ListForbiden = "";
    $FlagForbiden = 0;
    
    if ( length($SpecFile) == 0 ) { $PercentLen = 100; }
    else { $PercentLen = 100 - $MinLen; }

    $buf = $FileRep.$SepDir."bak";
    if ( !(mkdir $buf) ) {
        $buf .= $SepDir."*";
        if ( $^O eq 'linux' ) { system("rm -f $buf"); }
        else { system("del /F /Q $buf"); }
        }

    $bAff = "\n Reading of aligned files\n";    
    if ( $FlagIHM == 1 ) { ViewDetails($bAff, 1); }
    else { print $bAff; }
    foreach my $F (sort @T1) {
        if ( $FlagVerb == 1 ) {
            if ( $FlagIHM == 1 ) { ViewDetails("   === $F ===\n", 1); }
            else { print "   === $F ===\n"; }
            }
        $LSpe = "<<<<<";      # liste des sequences du fichier
        $SpeOut = "";    # liste des sequences avec noms identiques
        $Lg = 0;
        @T2 = ();
        @NewFic = ();    # stockage fichier avec nom de sequences modifies
        $n = 0;
        @tName = ();     # liste des noms de sequences modifies
        $SeqName = "";
        $Seq = "";
        if ( !open(FIC, $FileRep.$SepDir.$F) )
            { AffError("Can\'t open file \'$FileRep$F\'"); }
        $FlagDiese = 0;
        $ListForbiden .= "$F\n";
        push @NewFic, "#\n#\n";
        $flag = 0;
        while ( $L = <FIC> ) {
            #--- recupere le nom de l espece -------------------
            $L = EndLine($L);
            # ligne vide ou commentaire
            if ( length($L) == 0 or $L =~ m/^#/ ) { next; }
            # supprime commentaire en fin de ligne
            if ( $L =~ m/(.*)#(.+)/ ) { $FlagDiese = 1; $L = $1; }
            # ligne de nom de sequence
            if ( $L =~ m/^>(.*)/ ) {
                #--- recupere le nom de l'espece suivante ---
                $E = $1;
                chomp($E);
                $E =~ s/\s*$//;
                $E =~ s/^\s*//;
                #-- stocke sequence precedente ---
                if ( $SeqName ne '' ) {
                    # verifie que le type de sequence est toujours le meme
                    if ( $flag == 0 ) {
                        $L2 = $Seq;
                        $L2 =~ s/ //g;
                        if ( length($L2) > 0 ) { 
                            $TypeSeq = TestSeq($TypeSeq, $Seq); 
                            $flag = 1;
                            }
                        }
                    # sauvegarde pour reecriture du fichier
                    push @NewFic, ">$SeqName\n$Seq\n";
                    # determine la longueur maximale des sequences
                    $Lg = length($Seq) > $Lg ? length($Seq) : $Lg;
                    # compte les 'non-caracteres'
                    $L2 = $Seq;
                    $L2 =~ s/ |X|-|\*|\.|\$/\?/g;
                    $nbBlancs = 0 ;
                    $nbBlancs++ while ( $L2 =~ m/\?/g );
                    $nbBlancs++;
                    # verifie si nouvelle espece
                    $b1 = SearchSpecName($SeqName, "@");
                    push (@T2, $b1);
                    $b = "<<<<<" . $b1 . "<<<<<";
                    if ( !($ListEsp =~ m/$b/)) {
                        $ListEsp .= ($b1 . "<<<<<");
                        # remplit un hachage
                        # avec cle = nom espece / valeur = taux de presence
                        $FreqSpec{$b1} = 0;
                        $NbEspTot++;
                        }
                    }
                #-- reinitialise variables ---
                $SeqName = "";
                $Seq = "";

                #-- verifie si nom contient des caracteres interdits  ---
                if ( ForbidenChar($E) ) {
                    $ListForbiden .= "    ".$E."\n"; 
                    $FlagForbiden = 1;
                    }
                else {
                    #-- verifie si nom de sequence en double ---
                    $buf = "<<<<<" . $E . "<<<<<";
                    $buf =~ s/\|/!:!/g;        # identifiants de sequences avec |
                    if ( $LSpe =~ m/$buf/ ) {
                        if ( $Choose eq "" ) {
                            $bAff = "Many sequences have the same name\nDo you want modify name of sequences ?";
                            if ( $FlagIHM == 1 ) {
                                $Choose = AskModif($bAff);
                                }
                            else {
                                print "\n  $bAff (y/n) ";
                                $L = <STDIN>;
                                $L =~ s/\s+//g;
                                $Choose = $L;
                                }
                            }
                        if ( $Choose =~ m/yes|y/i ) {
                            # sauvegarde pour reecriture du fichier
                            $SeqName = $E."_".++$n;
                            push @tName, $SeqName;
                            }
                        else {
                            # liste des especes avec nom identique
                            if ( !($SpeOut =~ m/$E/) ) { $SpeOut .= "      $E\n"; }
                            }
                        }
                    else {
                        $SeqName = $E;
                        }
                    $LSpe .= ($E."<<<<<");
                    }
                }
            
            #--- recupere la sequence --------------------------
            elsif ( length($SeqName) > 0 )
                { $Seq .= $L; }
            }
        #-- stocke derniere sequence ---
        if ( $SeqName ne '' ) {
            # sauvegarde pour reecriture du fichier
            push @NewFic, ">$SeqName\n$Seq\n";
            # determine la longueur maximale des sequences
            $Lg = length($Seq) > $Lg ? length($Seq) : $Lg;
            # compte les 'non-caracteres'
            $L2 = $Seq;
            $L2 =~ s/ |X|-|\*|\.|\$/\?/g;
            $nbBlancs = 0 ;
            $nbBlancs++ while ( $L2 =~ m/\?/g );
            $nbBlancs++;
            # verifie si la sequence est assez longue
            if ( int($nbBlancs*100/$Lg) <= $PercentLen ) {
                # verifie si nouvelle espece
                $b1 = SearchSpecName($SeqName, "@");
                push (@T2, $b1);
                $b = "<<<<<" . $b1 . "<<<<<";
                if ( !($ListEsp =~ m/$b/)) {
                    $ListEsp .= ($b1 . "<<<<<");
                    # remplit un hachage
                    # avec cle = nom espece / valeur = taux de presence
                    $FreqSpec{$b1} = 0;
                    $NbEspTot++;
                    }
                }
            }
        close FIC;
        
        # avertissement si ligne coupee apres #
        if ( $FlagDiese == 1 ) {
            $bAff = "Lines with hash character (#) within them have been cut in file \'$FileRep$F\'";
            if ( $FlagIHM == 1 ) { ViewInfo('Warning', $bAff, 'blue'); }
            else { print "\n   WARNING : $bAff\n\n"; }
            }
        # arret si caracteres interdits
        if ( $FlagForbiden == 1 )
            { AffError("Some files contain sequence names with forbiden characters \n \\ \| \( \) \[ \] \{ \} \^ \$ \* \+ \? ~ ! % & = ' ; : , < > ]\n".$ListForbiden); }
        
        # programme arrete si au moins 1 fichier avec noms d'especes identiques
        if ( length($SpeOut) > 0 ) {
            $FicOut .= "   $F\n$SpeOut\n";
            }
        # remplit un hachage de tableaux
        # avec cle = nom du fichier / valeur = tableau des especes
        elsif ( ($FlagFile == 0) and ($FlagAll == 0) ) {
            $Fichiers{$F} = [ sort @T2 ];
            $FicLen{$F} = $Lg;
            $NbFileAli++;
            # reecrit le fichier
            $buf = $FileRep.$SepDir.$F;
            if ( $^O eq 'linux' ) { system("cp $buf ".$FileRep.$SepDir."bak"); }
            else { system("copy /Y $buf ".$FileRep.$SepDir."bak"); }
            if ( !open(FIC, "> ".$FileRep.$SepDir.$F) )
                { AffError("Can\'t open file \'$FileRep".'/'."$F\'"); }
            print FIC @NewFic;
            close FIC;
            # sauvegarde noms de sequences qui ont change
            if ( $#tName > 0 ) {
                $FicOut .= "\n--- $F ---\n";
                for ( sort @tName ) { $FicOut .= "   $_\n"; }
                }
            }
        }

    # programme arrete si au moins 1 fichier avec noms d'especes identiques
    if ( length($FicOut) > 0 ) {
        $L = $NameCef.".mod.log";
        if ( !open(LOG,"> $L") )
            { AffError("Can\'t open file \'$L\'"); }
        if ( $Choose =~ m/yes|y/i ) {
            print LOG "\n\nList of sequence's names modified :\n\n$FicOut\n\n\n";
            $L = "\n\n List of sequence's names modified in file \'$L\'\n\n";
            if ( $FlagIHM == 1 ) { ViewDetails($L, 0); }
            else { print $L; }    
            }
        else {
            print LOG "\n\nFiles with at least 2 sequences with the same name !\n\n$FicOut\n\n\n";
            AffError("\n\n Files with at least 2 sequences with the same name !\n\nList in file \'$L\'\n\n\n");
            }
        close LOG;
        }
}


1;

### fin du fichier                                           ###
################################################################
